@@ -1,5 +1,6 @@
 import asyncio
 import uuid
+from collections.abc import AsyncIterator
 from datetime import datetime, timezone
 from enum import Enum
 from typing import Any, Literal, Optional, Union
@@ -7,6 +8,8 @@ from typing import Any, Literal, Optional, Union
 from pydantic import AnyUrl, BaseModel, ConfigDict, Field
 
 from acp_sdk.models.errors import ACPError, Error
+from acp_sdk.models.types import AgentName, ResourceId, ResourceUrl, RunId, SessionId
+from acp_sdk.shared import ResourceLoader, ResourceStore
 
 
 class AnyModel(BaseModel):
@@ -74,6 +77,40 @@ class Metadata(BaseModel):
     model_config = ConfigDict(extra="allow")
 
 
+class CitationMetadata(BaseModel):
+    """
+    Represents an inline citation, providing info about information source. This
+    is supposed to be rendered as an inline icon, optionally marking a text
+    range it belongs to.
+
+    If CitationMetadata is included together with content in the message part,
+    the citation belongs to that content and renders at the MessagePart position.
+    This way may be used for non-text content, like images and files.
+
+    Alternatively, `start_index` and `end_index` may define a text range,
+    counting characters in the current Message across all MessageParts with
+    content type `text/*`, where the citation will be rendered. If one of
+    `start_index` and `end_index` is missing or their values are equal, the
+    citation renders only as an inline icon at that position.
+
+    If both `start_index` and `end_index` are not present and MessagePart has no
+    content, the citation renders as inline icon only at the MessagePart position.
+
+    Properties:
+    - url: URL of the source document.
+    - title: Title of the source document.
+    - description: Accompanying text, which may be a general description of the
+                   source document, or a specific snippet.
+    """
+
+    kind: Literal["citation"] = "citation"
+    start_index: Optional[int]
+    end_index: Optional[int]
+    url: Optional[str]
+    title: Optional[str]
+    description: Optional[str]
+
+
 class MessagePart(BaseModel):
     name: Optional[str] = None
     content_type: Optional[str] = "text/plain"
@@ -83,9 +120,9 @@ class MessagePart(BaseModel):
 
     model_config = ConfigDict(extra="allow")
 
+    metadata: Optional[CitationMetadata] = Field(discriminator="kind", default=None)
+
     def model_post_init(self, __context: Any) -> None:
-        if self.content is None and self.content_url is None:
-            raise ValueError("Either content or content_url must be provided")
         if self.content is not None and self.content_url is not None:
             raise ValueError("Only one of content or content_url can be provided")
 
@@ -95,6 +132,7 @@ class Artifact(MessagePart):
 
 
 class Message(BaseModel):
+    role: Literal["user"] | Literal["agent"] | str = Field("user", pattern=r"^(user|agent(\/[a-zA-Z0-9_\-]+)?)$")
     parts: list[MessagePart]
     created_at: datetime | None = Field(default_factory=lambda: datetime.now(timezone.utc))
     completed_at: datetime | None = Field(default_factory=lambda: datetime.now(timezone.utc))
@@ -102,7 +140,10 @@ class Message(BaseModel):
     def __add__(self, other: "Message") -> "Message":
         if not isinstance(other, Message):
             raise TypeError(f"Cannot concatenate Message with {type(other).__name__}")
+        if self.role != other.role:
+            raise ValueError("Cannot concatenate messages with different roles")
         return Message(
+            role=self.role,
             parts=self.parts + other.parts,
             created_at=min(self.created_at, other.created_at) if self.created_at and other.created_at else None,
             completed_at=max(self.completed_at, other.completed_at)
@@ -146,11 +187,6 @@ class Message(BaseModel):
         return Message(parts=parts, created_at=self.created_at, completed_at=self.completed_at)
 
 
-AgentName = str
-SessionId = uuid.UUID
-RunId = uuid.UUID
-
-
 class RunMode(str, Enum):
     SYNC = "sync"
     ASYNC = "async"
@@ -280,11 +316,41 @@ Event = Union[
     RunCancelledEvent,
     RunFailedEvent,
     RunCompletedEvent,
-    MessagePartEvent,
 ]
 
 
-class Agent(BaseModel):
+class AgentManifest(BaseModel):
     name: str
     description: str | None = None
     metadata: Metadata = Metadata()
+
+
+class Session(BaseModel):
+    id: SessionId = Field(default_factory=uuid.uuid4)
+    history: list[ResourceUrl] = Field(default_factory=list)
+    state: ResourceUrl | None = None
+
+    loader: ResourceLoader | None = Field(None, exclude=True)
+    store: ResourceStore | None = Field(None, exclude=True)
+
+    model_config = ConfigDict(arbitrary_types_allowed=True)
+
+    async def load_history(self, *, loader: ResourceLoader | None = None) -> AsyncIterator[Message]:
+        loader = loader or self.loader or ResourceLoader()
+        for url in self.history:
+            data = await loader.load(url)
+            yield Message.model_validate_json(data)
+
+    async def load_state(self, *, loader: ResourceLoader | None = None) -> bytes:
+        loader = loader or self.loader or ResourceLoader()
+        data = await loader.load(self.state)
+        return data
+
+    async def store_state(self, data: bytes, *, store: ResourceStore | None = None) -> ResourceUrl:
+        store = store or self.store
+        if not store:
+            raise ValueError("Store must be specified")
+
+        id = ResourceId()
+        await store.store(id, data)
+        return await store.url(id)
