@@ -1,5 +1,6 @@
 import asyncio
 import os
+import re
 from collections.abc import AsyncGenerator, Awaitable
 from contextlib import asynccontextmanager
 from typing import Any, Callable
@@ -10,7 +11,7 @@ import uvicorn.config
 from fastapi import FastAPI
 
 from acp_sdk.models import Metadata
-from acp_sdk.server.agent import Agent
+from acp_sdk.server.agent import AgentManifest
 from acp_sdk.server.agent import agent as agent_decorator
 from acp_sdk.server.app import create_app
 from acp_sdk.server.logging import configure_logger as configure_logger_func
@@ -18,11 +19,12 @@ from acp_sdk.server.logging import logger
 from acp_sdk.server.store import Store
 from acp_sdk.server.telemetry import configure_telemetry as configure_telemetry_func
 from acp_sdk.server.utils import async_request_with_retry
+from acp_sdk.shared.resources import ResourceLoader, ResourceStore
 
 
 class Server:
     def __init__(self) -> None:
-        self.agents: list[Agent] = []
+        self.agents: list[AgentManifest] = []
         self.server: uvicorn.Server | None = None
 
     def agent(
@@ -41,7 +43,7 @@ class Server:
 
         return decorator
 
-    def register(self, *agents: Agent) -> None:
+    def register(self, *agents: AgentManifest) -> None:
         self.agents.extend(agents)
 
     @asynccontextmanager
@@ -55,6 +57,8 @@ class Server:
         configure_telemetry: bool = False,
         self_registration: bool = True,
         store: Store | None = None,
+        resource_store: ResourceStore | None = None,
+        resource_loader: ResourceLoader | None = None,
         host: str = "127.0.0.1",
         port: int = 8000,
         uds: str | None = None,
@@ -117,7 +121,13 @@ class Server:
 
         import uvicorn
 
-        app = create_app(*self.agents, lifespan=self.lifespan, store=store)
+        app = create_app(
+            *self.agents,
+            lifespan=self.lifespan,
+            store=store,
+            resource_loader=resource_loader,
+            resource_store=resource_store,
+        )
 
         if configure_logger:
             configure_logger_func()
@@ -184,6 +194,8 @@ class Server:
         configure_telemetry: bool = False,
         self_registration: bool = True,
         store: Store | None = None,
+        resource_store: ResourceStore | None = None,
+        resource_loader: ResourceLoader | None = None,
         host: str = "127.0.0.1",
         port: int = 8000,
         uds: str | None = None,
@@ -242,6 +254,8 @@ class Server:
                 configure_telemetry=configure_telemetry,
                 self_registration=self_registration,
                 store=store,
+                resource_store=resource_store,
+                resource_loader=resource_loader,
                 host=host,
                 port=port,
                 uds=uds,
@@ -313,22 +327,33 @@ class Server:
             return
 
         url = os.getenv("PLATFORM_URL", "http://127.0.0.1:8333")
+        host = re.sub(r"localhost|127\.0\.0\.1", "host.docker.internal", self.server.config.host)
         request_data = {
-            "location": f"http://{self.server.config.host}:{self.server.config.port}",
+            "location": f"http://{host}:{self.server.config.port}",
         }
+        await async_request_with_retry(lambda client, data=request_data: client.get(f"{url}/api/v1/providers"))
         try:
             await async_request_with_retry(
-                lambda client, data=request_data: client.post(f"{url}/api/v1/providers/register/unmanaged", json=data)
+                lambda client, data=request_data: client.post(
+                    f"{url}/api/v1/providers", json=data, params={"auto_remove": True}
+                )
             )
             logger.info("Agent registered to the beeai server.")
 
             # check missing env keyes
             envs_request = await async_request_with_retry(lambda client: client.get(f"{url}/api/v1/variables"))
             envs = envs_request.get("env")
+            os.environ["LLM_MODEL"] = "dummy"
+            os.environ["LLM_API_KEY"] = "dummy"
+            os.environ["LLM_API_BASE"] = f"{url.rstrip('/')}/api/v1/llm"
+
             for agent in self.agents:
                 # register all available envs
                 missing_keyes = []
                 for env in agent.metadata.model_dump().get("env", []):
+                    # Those envs are set to use LLM gateway from platform server
+                    if env["name"] in {"LLM_MODEL", "LLM_API_KEY", "LLM_API_BASE"}:
+                        continue
                     server_env = envs.get(env.get("name"))
                     if server_env:
                         logger.debug(f"Env variable {env['name']} = '{server_env}' added dynamically")
